//  RW Foundations
// ================
// The RW Foundational Libraries

// Filename: ./include/incls_all.hpp
// -The following lines pull in the project's real *.hpp files.
// -note: this file also acts as a location for a few utility features

#pragma once

#include <cstdio>
#include <iostream>
#include <memory>
#include <sstream>
#include <stdexcept>
#include <string>

using std::cerr;
using std::cout;
using std::exception;
using std::make_unique;
using std::ostringstream;
using std::printf;
using std::puts;
using std::runtime_error;
using std::string;
using std::unique_ptr;

//
// NOTE: It is probably a good idea to be cautious about moving things around
// hereafter within this file; some dependencies here are sensitive to ordering.
//

//------------------------------------------------------------------------------
/** Our alias for a @c void*
 */
using Address = void*;

//------------------------------------------------------------------------------
/** Treat an @c Address as a reference to a general type @a T
 * @param[in] pt the type's @c Address to take the reference to
 * @return @a pt 's reference
 */
template <class T>
T& reference_to(Address pt) {
  return *static_cast<T*>(pt);
}

//------------------------------------------------------------------------------
// THE RW 'SWITCHBOARD':
//

// clang-format off
namespace rw                       { class RW; }

namespace rw::animeyes             { class Animeyes; }
namespace rw::animeyes             { class Animeyes_b; }
namespace rw::arigato              { class Arigato; }
namespace rw::arigato              { class Arigato_b; }
namespace rw::atashi               { class Atashi; }
namespace rw::atashi               { class Atashi_b; }
namespace rw::bones                { class Bones; }
namespace rw::bones                { class Bones_b; }
namespace rw::chibitsu             { class Chibitsu; }
namespace rw::chibitsu             { class Chibitsu_b; }
namespace rw::curator              { class Curator; }
namespace rw::curator              { class Curator_b; }
namespace rw::curator::bumpmaster  { class Bumpmaster; }
namespace rw::curator::bumpmaster  { class Bumpmaster_b; }
namespace rw::curator::waifusearch { class Waifusearch; }
namespace rw::curator::waifusearch { class Waifusearch_b; }
namespace rw::electro              { class Electro; }
namespace rw::electro              { class Electro_b; }
namespace rw::gears                { class Gears; }
namespace rw::gears                { class Gears_b; }
namespace rw::graphicom            { class Graphicom; }
namespace rw::graphicom            { class Graphicom_b; }
namespace rw::hearsay              { class Hearsay; }
namespace rw::hearsay              { class Hearsay_b; }
namespace rw::shell                { class Shell; }
namespace rw::shell                { class Shell_b; }
namespace rw::shell::face          { class Face; }
namespace rw::shell::face          { class Face_b; }
namespace rw::wired                { class Wired; }
namespace rw::wired                { class Wired_b; }
// clang-format on

using rw::RW;

using rw::animeyes::Animeyes;
using rw::animeyes::Animeyes_b;
using rw::arigato::Arigato;
using rw::arigato::Arigato_b;
using rw::atashi::Atashi;
using rw::atashi::Atashi_b;
using rw::bones::Bones;
using rw::bones::Bones_b;
using rw::chibitsu::Chibitsu;
using rw::chibitsu::Chibitsu_b;
using rw::curator::Curator;
using rw::curator::Curator_b;
using rw::curator::bumpmaster::Bumpmaster;
using rw::curator::bumpmaster::Bumpmaster_b;
using rw::curator::waifusearch::Waifusearch;
using rw::curator::waifusearch::Waifusearch_b;
using rw::electro::Electro;
using rw::electro::Electro_b;
using rw::gears::Gears;
using rw::gears::Gears_b;
using rw::graphicom::Graphicom;
using rw::graphicom::Graphicom_b;
using rw::hearsay::Hearsay;
using rw::hearsay::Hearsay_b;
using rw::shell::Shell;
using rw::shell::Shell_b;
using rw::shell::face::Face;
using rw::shell::face::Face_b;
using rw::wired::Wired;
using rw::wired::Wired_b;

//------------------------------------------------------------------------------
/** Provides an identification structure for instantiated RW objects
 * @note: @b T ag, @b I d, @b P ointer
 *
 * This is simply a tagging-mechanism intended to be suitable for understanding
 * the potentially highly-complex callgraphs creatable inside deep RW (and
 * other) systems. IE, @a 'who-called-whom-and-who-answered-the-call?'.
 */
struct Tip {
  std::string   tag{"nobody"};  // sentinel for an unassigned
  unsigned long id{9001'42};    //  "
  RW*           ptr{nullptr};   //  "

  friend bool operator==(Tip const& tip_a, Tip const& tip_b);
};

//------------------------------------------------------------------------------
/** Prints a Tip item's elements out to the console, with an optional message
 * @param[in] tip the Tip data struct to print
 * @param[in] msg optional description (&tc) string
 */
void prt_tag(Tip const& tip, std::string const& msg = "");

//------------------------------------------------------------------------------

inline void error(std::string const& s) {
  ostringstream ss{};
  ss << s << '\n';
  throw runtime_error(ss.str());
}

inline void error(std::string const& s, int const i) {
  ostringstream ss{};
  ss << s << ": " << i << '\n';
  throw runtime_error(ss.str());
}

inline void error(std::string const& s1, std::string const& s2) {
  ostringstream ss{};
  ss << s1 << ": " << s2 << '\n';
  throw runtime_error(ss.str());
}

inline void error(std::string const& s1,
                  Tip const&         t,
                  std::string const& s2 = "") {
  ostringstream ss{};
  ss << s1 << " ( from " << t.tag << '.' << t.id << ", via " << s2 << " )\n";
  throw runtime_error(ss.str());
}

//------------------------------------------------------------------------------
#include "rw/rw_incls_all.hpp"  // spcfc. #include'd before the other rw classes
//

#include "animeyes/animeyes_incls_all.hpp"
#include "arigato/arigato_incls_all.hpp"
#include "atashi/atashi_incls_all.hpp"
#include "bones/bones_incls_all.hpp"
#include "chibitsu/chibitsu_incls_all.hpp"
#include "curator/curator_incls_all.hpp"
#include "electro/electro_incls_all.hpp"
#include "gears/gears_incls_all.hpp"
#include "graphicom/graphicom_incls_all.hpp"
#include "hearsay/hearsay_incls_all.hpp"
#include "shell/shell_incls_all.hpp"
#include "wired/wired_incls_all.hpp"

//    --=^=--

// "So is it true Sempai? Are robowaifus really better than any human woman?"

// rw_foundations_lib.v211124
// Copyright (2021)
// License MIT(Expat) https://opensource.org/licenses/MIT
